@@ -21,6 +21,7 @@ struct mrb_state;
 #define MRB_EACH_OBJ_BREAK 1
 typedef int (mrb_each_object_callback)(struct mrb_state *mrb, struct RBasic *obj, void *data);
 void mrb_objspace_each_objects(struct mrb_state *mrb, mrb_each_object_callback *callback, void *data);
+size_t mrb_objspace_page_slot_size(void);
 MRB_API void mrb_free_context(struct mrb_state *mrb, struct mrb_context *c);
 
 #ifndef MRB_GC_ARENA_SIZE
@@ -86,6 +87,8 @@ typedef struct mrb_gc {
 MRB_API mrb_bool
 mrb_object_dead_p(struct mrb_state *mrb, struct RBasic *object);
 
+#define MRB_GC_RED 7
+
 MRB_END_DECL
 
 #endif  /* MRUBY_GC_H */
