@@ -8,20 +8,14 @@
 # define _CRT_NONSTDC_NO_DEPRECATE
 #endif
 
-#ifndef MRB_WITHOUT_FLOAT
-#include <float.h>
-#include <math.h>
-#endif
-#include <limits.h>
-#include <stddef.h>
-#include <stdlib.h>
-#include <string.h>
 #include <mruby.h>
 #include <mruby/array.h>
 #include <mruby/class.h>
 #include <mruby/range.h>
 #include <mruby/string.h>
 #include <mruby/numeric.h>
+#include <mruby/presym.h>
+#include <string.h>
 
 typedef struct mrb_shared_string {
   int refcnt;
@@ -31,7 +25,7 @@ typedef struct mrb_shared_string {
 
 const char mrb_digitmap[] = "0123456789abcdefghijklmnopqrstuvwxyz";
 
-#define mrb_obj_alloc_string(mrb) ((struct RString*)mrb_obj_alloc((mrb), MRB_TT_STRING, (mrb)->string_class))
+#define mrb_obj_alloc_string(mrb) MRB_OBJ_ALLOC((mrb), MRB_TT_STRING, (mrb)->string_class)
 
 static struct RString*
 str_init_normal_capa(mrb_state *mrb, struct RString *s,
@@ -140,21 +134,6 @@ str_new(mrb_state *mrb, const char *p, size_t len)
   return str_init_normal(mrb, mrb_obj_alloc_string(mrb), p, len);
 }
 
-static inline void
-str_with_class(struct RString *s, mrb_value obj)
-{
-  s->c = mrb_str_ptr(obj)->c;
-}
-
-static mrb_value
-mrb_str_new_empty(mrb_state *mrb, mrb_value str)
-{
-  struct RString *s = str_new(mrb, 0, 0);
-
-  str_with_class(s, str);
-  return mrb_obj_value(s);
-}
-
 MRB_API mrb_value
 mrb_str_new_capa(mrb_state *mrb, size_t capa)
 {
@@ -175,19 +154,6 @@ mrb_str_new_capa(mrb_state *mrb, size_t capa)
   return mrb_obj_value(s);
 }
 
-#ifndef MRB_STR_BUF_MIN_SIZE
-# define MRB_STR_BUF_MIN_SIZE 128
-#endif
-
-MRB_API mrb_value
-mrb_str_buf_new(mrb_state *mrb, size_t capa)
-{
-  if (capa < MRB_STR_BUF_MIN_SIZE) {
-    capa = MRB_STR_BUF_MIN_SIZE;
-  }
-  return mrb_str_new_capa(mrb, capa);
-}
-
 static void
 resize_capa(mrb_state *mrb, struct RString *s, size_t capacity)
 {
@@ -271,7 +237,7 @@ str_modify_keep_ascii(mrb_state *mrb, struct RString *s)
 static void
 check_null_byte(mrb_state *mrb, mrb_value str)
 {
-  mrb_to_str(mrb, str);
+  mrb_ensure_string_type(mrb, str);
   if (memchr(RSTRING_PTR(str), '\0', RSTRING_LEN(str))) {
     mrb_raise(mrb, E_ARGUMENT_ERROR, "string contains null byte");
   }
@@ -297,10 +263,12 @@ static const char utf8len_codepage[256] =
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
-  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
+  1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
   3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,4,4,4,4,4,1,1,1,1,1,1,1,1,1,1,1,
 };
 
+#define utf8_islead(c) ((unsigned char)((c)&0xc0) != 0x80)
+
 mrb_int
 mrb_utf8len(const char* p, const char* e)
 {
@@ -312,7 +280,7 @@ mrb_utf8len(const char* p, const char* e)
   if (len == 1) return 1;
   if (len > e - p) return 1;
   for (i = 1; i < len; ++i)
-    if ((p[i] & 0xc0) != 0x80)
+    if (utf8_islead(p[i]))
       return 1;
   return len;
 }
@@ -320,15 +288,15 @@ mrb_utf8len(const char* p, const char* e)
 mrb_int
 mrb_utf8_strlen(const char *str, mrb_int byte_len)
 {
-  mrb_int total = 0;
+  mrb_int len = 0;
   const char *p = str;
   const char *e = p + byte_len;
 
   while (p < e) {
     p += mrb_utf8len(p, e);
-    total++;
+    len++;
   }
-  return total;
+  return len;
 }
 
 static mrb_int
@@ -583,9 +551,6 @@ str_share(mrb_state *mrb, struct RString *orig, struct RString *s)
   else if (RSTR_FSHARED_P(orig)) {
     str_init_fshared(orig, s, orig->as.heap.aux.fshared);
   }
-  else if (mrb_frozen_p(orig) && !RSTR_POOL_P(orig)) {
-    str_init_fshared(orig, s, orig);
-  }
   else {
     if (orig->as.heap.aux.capa > orig->as.heap.len) {
       orig->as.heap.ptr = (char *)mrb_realloc(mrb, orig->as.heap.ptr, len+1);
@@ -596,29 +561,6 @@ str_share(mrb_state *mrb, struct RString *orig, struct RString *s)
   }
 }
 
-mrb_value
-mrb_str_pool(mrb_state *mrb, const char *p, mrb_int len, mrb_bool nofree)
-{
-  struct RString *s = (struct RString *)mrb_malloc(mrb, sizeof(struct RString));
-
-  s->tt = MRB_TT_STRING;
-  s->c = mrb->string_class;
-  s->flags = 0;
-
-  if (RSTR_EMBEDDABLE_P(len)) {
-    str_init_embed(s, p, len);
-  }
-  else if (nofree) {
-    str_init_nofree(s, p, len);
-  }
-  else {
-    str_init_normal(mrb, s, p, len);
-  }
-  RSTR_SET_POOL_FLAG(s);
-  MRB_SET_FROZEN_FLAG(s);
-  return mrb_obj_value(s);
-}
-
 mrb_value
 mrb_str_byte_subseq(mrb_state *mrb, mrb_value str, mrb_int beg, mrb_int len)
 {
@@ -890,15 +832,22 @@ mrb_str_to_cstr(mrb_state *mrb, mrb_value str0)
 {
   struct RString *s;
 
+  const char *p = RSTRING_PTR(str0);
+  size_t len = RSTRING_LEN(str0);
   check_null_byte(mrb, str0);
-  s = str_new(mrb, RSTRING_PTR(str0), RSTRING_LEN(str0));
+  if (RSTR_EMBEDDABLE_P(len)) {
+    s = str_init_embed(mrb_obj_alloc_string(mrb), p, len);
+  }
+  else {
+    s = str_init_normal(mrb, mrb_obj_alloc_string(mrb), p, len);
+  }
   return RSTR_PTR(s);
 }
 
 MRB_API void
 mrb_str_concat(mrb_state *mrb, mrb_value self, mrb_value other)
 {
-  other = mrb_str_to_str(mrb, other);
+  other = mrb_obj_as_string(mrb, other);
   mrb_str_cat_str(mrb, self, other);
 }
 
@@ -948,14 +897,14 @@ static mrb_value
 mrb_str_size(mrb_state *mrb, mrb_value self)
 {
   mrb_int len = RSTRING_CHAR_LEN(self);
-  return mrb_fixnum_value(len);
+  return mrb_int_value(mrb, len);
 }
 
 static mrb_value
 mrb_str_bytesize(mrb_state *mrb, mrb_value self)
 {
   mrb_int len = RSTRING_LEN(self);
-  return mrb_fixnum_value(len);
+  return mrb_int_value(mrb, len);
 }
 
 /* 15.2.10.5.1  */
@@ -985,7 +934,6 @@ mrb_str_times(mrb_state *mrb, mrb_value self)
 
   len = RSTRING_LEN(self)*times;
   str2 = str_new(mrb, 0, len);
-  str_with_class(str2, self);
   p = RSTR_PTR(str2);
   if (len > 0) {
     n = RSTRING_LEN(self);
@@ -1070,7 +1018,7 @@ mrb_str_cmp_m(mrb_state *mrb, mrb_value str1)
   else {
     result = mrb_str_cmp(mrb, str1, str2);
   }
-  return mrb_fixnum_value(result);
+  return mrb_int_value(mrb, result);
 }
 
 static mrb_bool
@@ -1111,29 +1059,11 @@ mrb_str_equal_m(mrb_state *mrb, mrb_value str1)
 }
 /* ---------------------------------- */
 
-MRB_API mrb_value
-mrb_str_to_str(mrb_state *mrb, mrb_value str)
-{
-  switch (mrb_type(str)) {
-  case MRB_TT_STRING:
-    return str;
-  case MRB_TT_SYMBOL:
-    return mrb_sym_str(mrb, mrb_symbol(str));
-  case MRB_TT_FIXNUM:
-    return mrb_fixnum_to_str(mrb, str, 10);
-  case MRB_TT_CLASS:
-  case MRB_TT_MODULE:
-    return mrb_mod_to_s(mrb, str);
-  default:
-    return mrb_convert_type(mrb, str, MRB_TT_STRING, "String", "to_s");
-  }
-}
-
 /* obslete: use RSTRING_PTR() */
 MRB_API const char*
 mrb_string_value_ptr(mrb_state *mrb, mrb_value str)
 {
-  str = mrb_str_to_str(mrb, str);
+  str = mrb_obj_as_string(mrb, str);
   return RSTRING_PTR(str);
 }
 
@@ -1141,7 +1071,7 @@ mrb_string_value_ptr(mrb_state *mrb, mrb_value str)
 MRB_API mrb_int
 mrb_string_value_len(mrb_state *mrb, mrb_value ptr)
 {
-  mrb_to_str(mrb, ptr);
+  mrb_ensure_string_type(mrb, ptr);
   return RSTRING_LEN(ptr);
 }
 
@@ -1151,7 +1081,6 @@ mrb_str_dup(mrb_state *mrb, mrb_value str)
   struct RString *s = mrb_str_ptr(str);
   struct RString *dup = str_new(mrb, 0, 0);
 
-  str_with_class(dup, str);
   return str_replace(mrb, dup, s);
 }
 
@@ -1173,14 +1102,17 @@ static enum str_convert_range
 str_convert_range(mrb_state *mrb, mrb_value str, mrb_value indx, mrb_value alen, mrb_int *beg, mrb_int *len)
 {
   if (!mrb_undef_p(alen)) {
-    *beg = mrb_int(mrb, indx);
-    *len = mrb_int(mrb, alen);
+    *beg = mrb_as_int(mrb, indx);
+    *len = mrb_as_int(mrb, alen);
     return STR_CHAR_RANGE;
   }
   else {
     switch (mrb_type(indx)) {
-      case MRB_TT_FIXNUM:
-        *beg = mrb_fixnum(indx);
+      default:
+        indx = mrb_ensure_int_type(mrb, indx);
+        /* fall through */
+      case MRB_TT_INTEGER:
+        *beg = mrb_integer(indx);
         *len = 1;
         return STR_CHAR_RANGE;
 
@@ -1191,16 +1123,6 @@ str_convert_range(mrb_state *mrb, mrb_value str, mrb_value indx, mrb_value alen,
         return STR_BYTE_RANGE_CORRECTED;
 
       case MRB_TT_RANGE:
-        goto range_arg;
-
-      default:
-        indx = mrb_to_int(mrb, indx);
-        if (mrb_fixnum_p(indx)) {
-          *beg = mrb_fixnum(indx);
-          *len = 1;
-          return STR_CHAR_RANGE;
-        }
-range_arg:
         *len = RSTRING_CHAR_LEN(str);
         switch (mrb_range_beg_len(mrb, indx, beg, len, *len, TRUE)) {
           case MRB_RANGE_OK:
@@ -1210,14 +1132,12 @@ range_arg:
           default:
             break;
         }
-
-        mrb_raise(mrb, E_TYPE_ERROR, "can't convert to Fixnum");
     }
   }
   return STR_OUT_OF_RANGE;
 }
 
-static mrb_value
+mrb_value
 mrb_str_aref(mrb_state *mrb, mrb_value str, mrb_value indx, mrb_value alen)
 {
   mrb_int beg, len;
@@ -1246,17 +1166,17 @@ mrb_str_aref(mrb_state *mrb, mrb_value str, mrb_value indx, mrb_value alen)
 /* 15.2.10.5.34 */
 /*
  *  call-seq:
- *     str[fixnum]                 => fixnum or nil
- *     str[fixnum, fixnum]         => new_str or nil
- *     str[range]                  => new_str or nil
- *     str[other_str]              => new_str or nil
- *     str.slice(fixnum)           => fixnum or nil
- *     str.slice(fixnum, fixnum)   => new_str or nil
- *     str.slice(range)            => new_str or nil
- *     str.slice(other_str)        => new_str or nil
+ *     str[int]                 => int or nil
+ *     str[int, int]            => new_str or nil
+ *     str[range]               => new_str or nil
+ *     str[other_str]           => new_str or nil
+ *     str.slice(int)           => int or nil
+ *     str.slice(int, int)      => new_str or nil
+ *     str.slice(range)         => new_str or nil
+ *     str.slice(other_str)     => new_str or nil
  *
- *  Element Reference---If passed a single <code>Fixnum</code>, returns the code
- *  of the character at that position. If passed two <code>Fixnum</code>
+ *  Element Reference---If passed a single <code>Integer</code>, returns the code
+ *  of the character at that position. If passed two <code>Integer</code>
  *  objects, returns a substring starting at the offset given by the first, and
  *  a length given by the second. If given a range, a substring containing
  *  characters at offsets given by the range is returned. In all three cases, if
@@ -1310,13 +1230,11 @@ str_replace_partial(mrb_state *mrb, mrb_value src, mrb_int pos, mrb_int end, mrb
   if (end > len) { end = len; }
 
   if (pos < 0 || pos > len) {
-    str_out_of_index(mrb, mrb_fixnum_value(pos));
+    str_out_of_index(mrb, mrb_int_value(mrb, pos));
   }
 
   replen = (mrb_nil_p(rep) ? 0 : RSTRING_LEN(rep));
-  newlen = replen + len - (end - pos);
-
-  if (newlen >= MRB_SSIZE_MAX || newlen < replen /* overflowed */) {
+  if (mrb_int_add_overflow(replen, len - (end - pos), &newlen) || newlen >= MRB_SSIZE_MAX) {
     mrb_raise(mrb, E_RUNTIME_ERROR, "string size too big");
   }
 
@@ -1429,8 +1347,7 @@ mrb_str_aset(mrb_state *mrb, mrb_value str, mrb_value indx, mrb_value alen, mrb_
 {
   mrb_int beg, len, charlen;
 
-  mrb_to_str(mrb, replace);
-
+  mrb_ensure_string_type(mrb, replace);
   switch (str_convert_range(mrb, str, indx, alen, &beg, &len)) {
     case STR_OUT_OF_RANGE:
     default:
@@ -1447,14 +1364,17 @@ mrb_str_aset(mrb_state *mrb, mrb_value str, mrb_value indx, mrb_value alen, mrb_
       str_range_to_bytes(str, &beg, &len);
       /* fall through */
     case STR_BYTE_RANGE_CORRECTED:
-      str_replace_partial(mrb, str, beg, beg + len, replace);
+      if (mrb_int_add_overflow(beg, len, &len)) {
+        mrb_raise(mrb, E_RUNTIME_ERROR, "string index too big");
+      }
+      str_replace_partial(mrb, str, beg, len, replace);
   }
 }
 
 /*
  * call-seq:
- *    str[fixnum] = replace
- *    str[fixnum, fixnum] = replace
+ *    str[int] = replace
+ *    str[int, int] = replace
  *    str[range] = replace
  *    str[other_str] = replace
  *
@@ -1802,26 +1722,43 @@ mrb_str_substr(mrb_state *mrb, mrb_value str, mrb_int beg, mrb_int len)
   return str_substr(mrb, str, beg, len);
 }
 
+/*
+ * 32 bit magic FNV-0 and FNV-1 prime
+ */
+#define FNV_32_PRIME ((uint32_t)0x01000193)
+#define FNV1_32_INIT ((uint32_t)0x811c9dc5)
+
 uint32_t
 mrb_str_hash(mrb_state *mrb, mrb_value str)
 {
-  /* 1-8-7 */
   struct RString *s = mrb_str_ptr(str);
-  mrb_int len = RSTR_LEN(s);
-  char *p = RSTR_PTR(s);
-  uint64_t key = 0;
+  const unsigned char *bp = (unsigned char*)RSTR_PTR(s); /* start of buffer */
+  const unsigned char *be = bp + RSTR_LEN(s);            /* beyond end of buffer */
+  uint32_t hval = FNV1_32_INIT;
 
-  while (len--) {
-    key = key*65599 + *p;
-    p++;
+  /*
+   * FNV-1 hash each octet in the buffer
+   */
+  while (bp < be) {
+    /* multiply by the 32 bit FNV magic prime mod 2^32 */
+#if defined(NO_FNV_GCC_OPTIMIZATION)
+    hval *= FNV_32_PRIME;
+#else
+    hval += (hval<<1) + (hval<<4) + (hval<<7) + (hval<<8) + (hval<<24);
+#endif
+
+    /* xor the bottom with the current octet */
+    hval ^= (uint32_t)*bp++;
   }
-  return (uint32_t)(key + (key>>5));
+
+  /* return our new hash value */
+  return hval;
 }
 
 /* 15.2.10.5.20 */
 /*
  * call-seq:
- *    str.hash   => fixnum
+ *    str.hash   => int
  *
  * Return a hash based on the string's length and content.
  */
@@ -1829,14 +1766,14 @@ static mrb_value
 mrb_str_hash_m(mrb_state *mrb, mrb_value self)
 {
   mrb_int key = mrb_str_hash(mrb, self);
-  return mrb_fixnum_value(key);
+  return mrb_int_value(mrb, key);
 }
 
 /* 15.2.10.5.21 */
 /*
  *  call-seq:
  *     str.include? other_str   => true or false
- *     str.include? fixnum      => true or false
+ *     str.include? int         => true or false
  *
  *  Returns <code>true</code> if <i>str</i> contains the given string or
  *  character.
@@ -1859,7 +1796,7 @@ mrb_str_include(mrb_state *mrb, mrb_value self)
 /* 15.2.10.5.22 */
 /*
  *  call-seq:
- *     str.index(substring [, offset])   => fixnum or nil
+ *     str.index(substring [, offset])   => int or nil
  *
  *  Returns the index of the first occurrence of the given
  *  <i>substring</i>. Returns <code>nil</code> if not found.
@@ -1891,7 +1828,7 @@ mrb_str_index_m(mrb_state *mrb, mrb_value str)
 
   if (pos == -1) return mrb_nil_value();
   BYTES_ALIGN_CHECK(pos);
-  return mrb_fixnum_value(pos);
+  return mrb_int_value(mrb, pos);
 }
 
 /* 15.2.10.5.24 */
@@ -1940,7 +1877,7 @@ mrb_str_init(mrb_state *mrb, mrb_value self)
  *     str.to_sym   => symbol
  *
  *  Returns the <code>Symbol</code> corresponding to <i>str</i>, creating the
- *  symbol if it did not previously exist. See <code>Symbol#id2name</code>.
+ *  symbol if it did not previously exist.
  *
  *     "Koala".intern         #=> :Koala
  *     s = 'cat'.to_sym       #=> :cat
@@ -1962,10 +1899,20 @@ mrb_str_intern(mrb_state *mrb, mrb_value self)
 MRB_API mrb_value
 mrb_obj_as_string(mrb_state *mrb, mrb_value obj)
 {
-  if (mrb_string_p(obj)) {
+  switch (mrb_type(obj)) {
+  case MRB_TT_STRING:
     return obj;
+  case MRB_TT_SYMBOL:
+    return mrb_sym_str(mrb, mrb_symbol(obj));
+  case MRB_TT_INTEGER:
+    return mrb_integer_to_str(mrb, obj, 10);
+  case MRB_TT_SCLASS:
+  case MRB_TT_CLASS:
+  case MRB_TT_MODULE:
+    return mrb_mod_to_s(mrb, obj);
+  default:
+    return mrb_type_convert(mrb, obj, MRB_TT_STRING, MRB_SYM(to_s));
   }
-  return mrb_str_to_str(mrb, obj);
 }
 
 MRB_API mrb_value
@@ -2075,7 +2022,7 @@ mrb_str_reverse(mrb_state *mrb, mrb_value str)
 /* 15.2.10.5.31 */
 /*
  *  call-seq:
- *     str.rindex(substring [, offset])   => fixnum or nil
+ *     str.rindex(substring [, offset])   => int or nil
  *
  *  Returns the index of the last occurrence of the given <i>substring</i>.
  *  Returns <code>nil</code> if not found. If the second parameter is
@@ -2091,9 +2038,11 @@ static mrb_value
 mrb_str_rindex(mrb_state *mrb, mrb_value str)
 {
   mrb_value sub;
-  mrb_int pos, len = RSTRING_CHAR_LEN(str);
+  mrb_int pos;
+  int argc = mrb_get_args(mrb, "S|i", &sub, &pos);
+  mrb_int len = RSTRING_CHAR_LEN(str);
 
-  if (mrb_get_args(mrb, "S|i", &sub, &pos) == 1) {
+  if (argc == 1) {
     pos = len;
   }
   else {
@@ -2110,7 +2059,7 @@ mrb_str_rindex(mrb_state *mrb, mrb_value str)
   if (pos >= 0) {
     pos = bytes2chars(RSTRING_PTR(str), RSTRING_LEN(str), pos);
     BYTES_ALIGN_CHECK(pos);
-    return mrb_fixnum_value(pos);
+    return mrb_int_value(mrb, pos);
   }
   return mrb_nil_value();
 }
@@ -2239,7 +2188,7 @@ mrb_str_split_m(mrb_state *mrb, mrb_value str)
   }
   if (RSTRING_LEN(str) > 0 && (lim_p || RSTRING_LEN(str) > beg || lim < 0)) {
     if (RSTRING_LEN(str) == beg) {
-      tmp = mrb_str_new_empty(mrb, str);
+      tmp = mrb_str_new(mrb, 0, 0);
     }
     else {
       tmp = mrb_str_byte_subseq(mrb, str, beg, RSTRING_LEN(str)-beg);
@@ -2256,14 +2205,14 @@ mrb_str_split_m(mrb_state *mrb, mrb_value str)
   return result;
 }
 
-mrb_value
-mrb_str_len_to_inum(mrb_state *mrb, const char *str, mrb_int len, mrb_int base, int badcheck)
+static mrb_value
+mrb_str_len_to_integer(mrb_state *mrb, const char *str, size_t len, mrb_int base, int badcheck)
 {
   const char *p = str;
   const char *pend = str + len;
   char sign = 1;
   int c;
-  uint64_t n = 0;
+  mrb_int n = 0;
   mrb_int val;
 
 #define conv_digit(c) \
@@ -2390,19 +2339,17 @@ mrb_str_len_to_inum(mrb_state *mrb, const char *str, mrb_int len, mrb_int base,
     if (c < 0 || c >= base) {
       break;
     }
-    n *= base;
-    n += c;
-    if (n > (uint64_t)MRB_INT_MAX + (sign ? 0 : 1)) {
-#ifndef MRB_WITHOUT_FLOAT
-      if (base == 10) {
-        return mrb_float_value(mrb, mrb_str_to_dbl(mrb, mrb_str_new(mrb, str, len), badcheck));
-      }
-      else
-#endif
-      {
-        mrb_raisef(mrb, E_RANGE_ERROR, "string (%l) too big for integer", str, pend-str);
+    if (mrb_int_mul_overflow(n, base, &n)) goto overflow;
+    if (MRB_INT_MAX - c < n) {
+      if (sign == 0 && MRB_INT_MAX - n == c - 1) {
+        n = MRB_INT_MIN;
+        sign = 1;
+        break;
       }
+    overflow:
+      mrb_raisef(mrb, E_RANGE_ERROR, "string (%l) too big for integer", str, pend-str);
     }
+    n += c;
   }
   val = (mrb_int)n;
   if (badcheck) {
@@ -2412,7 +2359,7 @@ mrb_str_len_to_inum(mrb_state *mrb, const char *str, mrb_int len, mrb_int base,
     if (p<pend) goto bad;               /* trailing garbage */
   }
 
-  return mrb_fixnum_value(sign ? val : -val);
+  return mrb_int_value(mrb, sign ? val : -val);
  nullbyte:
   mrb_raise(mrb, E_ARGUMENT_ERROR, "string contains null byte");
   /* not reached */
@@ -2422,12 +2369,6 @@ mrb_str_len_to_inum(mrb_state *mrb, const char *str, mrb_int len, mrb_int base,
   return mrb_fixnum_value(0);
 }
 
-MRB_API mrb_value
-mrb_cstr_to_inum(mrb_state *mrb, const char *str, mrb_int base, mrb_bool badcheck)
-{
-  return mrb_str_len_to_inum(mrb, str, strlen(str), base, badcheck);
-}
-
 /* obslete: use RSTRING_CSTR() or mrb_string_cstr() */
 MRB_API const char*
 mrb_string_value_cstr(mrb_state *mrb, mrb_value *ptr)
@@ -2460,15 +2401,15 @@ mrb_string_cstr(mrb_state *mrb, mrb_value str)
 }
 
 MRB_API mrb_value
-mrb_str_to_inum(mrb_state *mrb, mrb_value str, mrb_int base, mrb_bool badcheck)
+mrb_str_to_integer(mrb_state *mrb, mrb_value str, mrb_int base, mrb_bool badcheck)
 {
   const char *s;
   mrb_int len;
 
-  mrb_to_str(mrb, str);
+  mrb_ensure_string_type(mrb, str);
   s = RSTRING_PTR(str);
   len = RSTRING_LEN(str);
-  return mrb_str_len_to_inum(mrb, s, len, base, badcheck);
+  return mrb_str_len_to_integer(mrb, s, len, base, badcheck);
 }
 
 /* 15.2.10.5.38 */
@@ -2498,14 +2439,14 @@ mrb_str_to_i(mrb_state *mrb, mrb_value self)
   mrb_int base = 10;
 
   mrb_get_args(mrb, "|i", &base);
-  if (base < 0) {
+  if (base < 0 || 36 < base) {
     mrb_raisef(mrb, E_ARGUMENT_ERROR, "illegal radix %i", base);
   }
-  return mrb_str_to_inum(mrb, self, base, FALSE);
+  return mrb_str_to_integer(mrb, self, base, FALSE);
 }
 
-#ifndef MRB_WITHOUT_FLOAT
-double
+#ifndef MRB_NO_FLOAT
+static double
 mrb_str_len_to_dbl(mrb_state *mrb, const char *s, size_t len, mrb_bool badcheck)
 {
   char buf[DBL_DIG * 4 + 20];
@@ -2525,9 +2466,9 @@ mrb_str_len_to_dbl(mrb_state *mrb, const char *s, size_t len, mrb_bool badcheck)
     mrb_value x;
 
     if (!badcheck) return 0.0;
-    x = mrb_str_len_to_inum(mrb, p, pend-p, 0, badcheck);
-    if (mrb_fixnum_p(x))
-      d = (double)mrb_fixnum(x);
+    x = mrb_str_len_to_integer(mrb, p, pend-p, 0, badcheck);
+    if (mrb_integer_p(x))
+      d = (double)mrb_integer(x);
     else /* if (mrb_float_p(x)) */
       d = mrb_float(x);
     return d;
@@ -2594,12 +2535,6 @@ bad:
   return d;
 }
 
-MRB_API double
-mrb_cstr_to_dbl(mrb_state *mrb, const char *s, mrb_bool badcheck)
-{
-  return mrb_str_len_to_dbl(mrb, s, strlen(s), badcheck);
-}
-
 MRB_API double
 mrb_str_to_dbl(mrb_state *mrb, mrb_value str, mrb_bool badcheck)
 {
@@ -2612,7 +2547,7 @@ mrb_str_to_dbl(mrb_state *mrb, mrb_value str, mrb_bool badcheck)
  *     str.to_f   => float
  *
  *  Returns the result of interpreting leading characters in <i>str</i> as a
- *  floating point number. Extraneous characters past the end of a valid number
+ *  floating-point number. Extraneous characters past the end of a valid number
  *  are ignored. If there is not a valid number at the start of <i>str</i>,
  *  <code>0.0</code> is returned. This method never raises an exception.
  *
@@ -2770,7 +2705,7 @@ mrb_str_cat_str(mrb_state *mrb, mrb_value str, mrb_value str2)
 MRB_API mrb_value
 mrb_str_append(mrb_state *mrb, mrb_value str1, mrb_value str2)
 {
-  mrb_to_str(mrb, str2);
+  mrb_ensure_string_type(mrb, str2);
   return mrb_str_cat_str(mrb, str1, str2);
 }
 
@@ -2793,7 +2728,7 @@ mrb_str_inspect(mrb_state *mrb, mrb_value str)
 
 /*
  * call-seq:
- *   str.bytes   -> array of fixnums
+ *   str.bytes   -> array of int
  *
  * Returns an array of bytes in _str_.
  *
@@ -2886,23 +2821,25 @@ static mrb_value
 mrb_str_byteslice(mrb_state *mrb, mrb_value str)
 {
   mrb_value a1;
-  mrb_int str_len = RSTRING_LEN(str), beg, len;
+  mrb_int str_len, beg, len;
   mrb_bool empty = TRUE;
 
   len = mrb_get_argc(mrb);
   switch (len) {
   case 2:
     mrb_get_args(mrb, "ii", &beg, &len);
+    str_len = RSTRING_LEN(str);
     break;
   case 1:
     a1 = mrb_get_arg1(mrb);
+    str_len = RSTRING_LEN(str);
     if (mrb_range_p(a1)) {
       if (mrb_range_beg_len(mrb, a1, &beg, &len, str_len, TRUE) != MRB_RANGE_OK) {
         return mrb_nil_value();
       }
     }
     else {
-      beg = mrb_fixnum(mrb_to_int(mrb, a1));
+      beg = mrb_as_int(mrb, a1);
       len = 1;
       empty = FALSE;
     }
@@ -2919,6 +2856,51 @@ mrb_str_byteslice(mrb_state *mrb, mrb_value str)
   }
 }
 
+static mrb_value
+sub_replace(mrb_state *mrb, mrb_value self)
+{
+  char *p, *match;
+  mrb_int plen, mlen;
+  mrb_int found, offset;
+  mrb_value result;
+
+  mrb_get_args(mrb, "ssi", &p, &plen, &match, &mlen, &found);
+  result = mrb_str_new(mrb, 0, 0);
+  for (mrb_int i=0; i<plen; i++) {
+    if (p[i] != '\\' || i+1==plen) {
+      mrb_str_cat(mrb, result, p+i, 1);
+      continue;
+    }
+    i++;
+    switch (p[i]) {
+    case '\\':
+      mrb_str_cat(mrb, result, "\\", 1);
+      break;
+    case '`':
+      mrb_str_cat(mrb, result, RSTRING_PTR(self), chars2bytes(self, 0, found));
+      break;
+    case '&': case '0':
+      mrb_str_cat(mrb, result, match, mlen);
+      break;
+    case '\'':
+      offset = chars2bytes(self, 0, found) + mlen;
+      if (RSTRING_LEN(self) > offset) {
+        mrb_str_cat(mrb, result, RSTRING_PTR(self)+offset, RSTRING_LEN(self)-offset);
+      }
+      break;
+    case '1': case '2': case '3':
+    case '4': case '5': case '6':
+    case '7': case '8': case '9':
+      /* ignore sub-group match (no Regexp supported) */
+      break;
+    default:
+      mrb_str_cat(mrb, result, &p[i-1], 2);
+      break;
+    }
+  }
+  return result;
+}
+
 /* ---------------------------*/
 void
 mrb_init_string(mrb_state *mrb)
@@ -2965,7 +2947,7 @@ mrb_init_string(mrb_state *mrb)
   mrb_define_method(mrb, s, "slice",           mrb_str_aref_m,          MRB_ARGS_ANY());  /* 15.2.10.5.34 */
   mrb_define_method(mrb, s, "split",           mrb_str_split_m,         MRB_ARGS_ANY());  /* 15.2.10.5.35 */
 
-#ifndef MRB_WITHOUT_FLOAT
+#ifndef MRB_NO_FLOAT
   mrb_define_method(mrb, s, "to_f",            mrb_str_to_f,            MRB_ARGS_NONE()); /* 15.2.10.5.38 */
 #endif
   mrb_define_method(mrb, s, "to_i",            mrb_str_to_i,            MRB_ARGS_ANY());  /* 15.2.10.5.39 */
@@ -2980,251 +2962,6 @@ mrb_init_string(mrb_state *mrb)
   mrb_define_method(mrb, s, "getbyte",         mrb_str_getbyte,         MRB_ARGS_REQ(1));
   mrb_define_method(mrb, s, "setbyte",         mrb_str_setbyte,         MRB_ARGS_REQ(2));
   mrb_define_method(mrb, s, "byteslice",       mrb_str_byteslice,       MRB_ARGS_ARG(1,1));
+
+  mrb_define_method(mrb, s, "__sub_replace",   sub_replace,             MRB_ARGS_REQ(3)); /* internal */
 }
-
-#ifndef MRB_WITHOUT_FLOAT
-/*
- * Source code for the "strtod" library procedure.
- *
- * Copyright (c) 1988-1993 The Regents of the University of California.
- * Copyright (c) 1994 Sun Microsystems, Inc.
- *
- * Permission to use, copy, modify, and distribute this
- * software and its documentation for any purpose and without
- * fee is hereby granted, provided that the above copyright
- * notice appear in all copies.  The University of California
- * makes no representations about the suitability of this
- * software for any purpose.  It is provided "as is" without
- * express or implied warranty.
- *
- * RCS: @(#) $Id: strtod.c 11708 2007-02-12 23:01:19Z shyouhei $
- */
-
-#include <ctype.h>
-#include <errno.h>
-
-static const int maxExponent = 511; /* Largest possible base 10 exponent.  Any
-                                     * exponent larger than this will already
-                                     * produce underflow or overflow, so there's
-                                     * no need to worry about additional digits.
-                                     */
-static const double powersOf10[] = {/* Table giving binary powers of 10.  Entry */
-    10.,                            /* is 10^2^i.  Used to convert decimal */
-    100.,                           /* exponents into floating-point numbers. */
-    1.0e4,
-    1.0e8,
-    1.0e16,
-    1.0e32,
-    1.0e64,
-    1.0e128,
-    1.0e256
-};
-
-MRB_API double
-mrb_float_read(const char *string, char **endPtr)
-/*  const char *string;            A decimal ASCII floating-point number,
-                                 * optionally preceded by white space.
-                                 * Must have form "-I.FE-X", where I is the
-                                 * integer part of the mantissa, F is the
-                                 * fractional part of the mantissa, and X
-                                 * is the exponent.  Either of the signs
-                                 * may be "+", "-", or omitted.  Either I
-                                 * or F may be omitted, or both.  The decimal
-                                 * point isn't necessary unless F is present.
-                                 * The "E" may actually be an "e".  E and X
-                                 * may both be omitted (but not just one).
-                                 */
-/*  char **endPtr;                 If non-NULL, store terminating character's
-                                 * address here. */
-{
-    int sign, expSign = FALSE;
-    double fraction, dblExp;
-    const double *d;
-    const char *p;
-    int c;
-    int exp = 0;                /* Exponent read from "EX" field. */
-    int fracExp = 0;            /* Exponent that derives from the fractional
-                                 * part.  Under normal circumstatnces, it is
-                                 * the negative of the number of digits in F.
-                                 * However, if I is very long, the last digits
-                                 * of I get dropped (otherwise a long I with a
-                                 * large negative exponent could cause an
-                                 * unnecessary overflow on I alone).  In this
-                                 * case, fracExp is incremented one for each
-                                 * dropped digit. */
-    int mantSize;               /* Number of digits in mantissa. */
-    int decPt;                  /* Number of mantissa digits BEFORE decimal
-                                 * point. */
-    const char *pExp;           /* Temporarily holds location of exponent
-                                 * in string. */
-
-    /*
-     * Strip off leading blanks and check for a sign.
-     */
-
-    p = string;
-    while (ISSPACE(*p)) {
-      p += 1;
-    }
-    if (*p == '-') {
-      sign = TRUE;
-      p += 1;
-    }
-    else {
-      if (*p == '+') {
-        p += 1;
-      }
-      sign = FALSE;
-    }
-
-    /*
-     * Count the number of digits in the mantissa (including the decimal
-     * point), and also locate the decimal point.
-     */
-
-    decPt = -1;
-    for (mantSize = 0; ; mantSize += 1)
-    {
-      c = *p;
-      if (!ISDIGIT(c)) {
-        if ((c != '.') || (decPt >= 0)) {
-          break;
-        }
-        decPt = mantSize;
-      }
-      p += 1;
-    }
-
-    /*
-     * Now suck up the digits in the mantissa.  Use two integers to
-     * collect 9 digits each (this is faster than using floating-point).
-     * If the mantissa has more than 18 digits, ignore the extras, since
-     * they can't affect the value anyway.
-     */
-
-    pExp  = p;
-    p -= mantSize;
-    if (decPt < 0) {
-      decPt = mantSize;
-    }
-    else {
-      mantSize -= 1; /* One of the digits was the point. */
-    }
-    if (mantSize > 18) {
-      if (decPt - 18 > 29999) {
-        fracExp = 29999;
-      }
-      else {
-        fracExp = decPt - 18;
-      }
-      mantSize = 18;
-    }
-    else {
-      fracExp = decPt - mantSize;
-    }
-    if (mantSize == 0) {
-      fraction = 0.0;
-      p = string;
-      goto done;
-    }
-    else {
-      int frac1, frac2;
-      frac1 = 0;
-      for ( ; mantSize > 9; mantSize -= 1)
-      {
-        c = *p;
-        p += 1;
-        if (c == '.') {
-          c = *p;
-          p += 1;
-        }
-        frac1 = 10*frac1 + (c - '0');
-      }
-      frac2 = 0;
-      for (; mantSize > 0; mantSize -= 1)
-      {
-        c = *p;
-        p += 1;
-        if (c == '.') {
-          c = *p;
-          p += 1;
-        }
-        frac2 = 10*frac2 + (c - '0');
-      }
-      fraction = (1.0e9 * frac1) + frac2;
-    }
-
-    /*
-     * Skim off the exponent.
-     */
-
-    p = pExp;
-    if ((*p == 'E') || (*p == 'e')) {
-      p += 1;
-      if (*p == '-') {
-        expSign = TRUE;
-        p += 1;
-      }
-      else {
-        if (*p == '+') {
-          p += 1;
-        }
-        expSign = FALSE;
-      }
-      while (ISDIGIT(*p)) {
-        exp = exp * 10 + (*p - '0');
-        if (exp > 19999) {
-          exp = 19999;
-        }
-        p += 1;
-      }
-    }
-    if (expSign) {
-      exp = fracExp - exp;
-    }
-    else {
-      exp = fracExp + exp;
-    }
-
-    /*
-     * Generate a floating-point number that represents the exponent.
-     * Do this by processing the exponent one bit at a time to combine
-     * many powers of 2 of 10. Then combine the exponent with the
-     * fraction.
-     */
-
-    if (exp < 0) {
-      expSign = TRUE;
-      exp = -exp;
-    }
-    else {
-      expSign = FALSE;
-    }
-    if (exp > maxExponent) {
-      exp = maxExponent;
-      errno = ERANGE;
-    }
-    dblExp = 1.0;
-    for (d = powersOf10; exp != 0; exp >>= 1, d += 1) {
-      if (exp & 01) {
-        dblExp *= *d;
-      }
-    }
-    if (expSign) {
-      fraction /= dblExp;
-    }
-    else {
-      fraction *= dblExp;
-    }
-
-done:
-    if (endPtr != NULL) {
-      *endPtr = (char *) p;
-    }
-
-    if (sign) {
-      return -fraction;
-    }
-    return fraction;
-}
-#endif
